/* Copyright (c) 2000-2012 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

xtk.include('domutils');
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

if (typeof(ko)=='undefined') {
    var ko = {};
}
if (typeof(ko.workspace2) == "undefined") {
    ko.workspace2 = {};
}


(function() {

var platform = require("sdk/system").platform;
var log = require("ko/logging").getLogger('workspace2');
//log.setLevel(ko.logging.LOG_DEBUG);

var _saveInProgress = false;
var _restoreInProgress = false;
var _bundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
    .getService(Components.interfaces.nsIStringBundleService)
    .createBundle("chrome://komodo/locale/views.properties");

this.saveInProgress = function() {
    return _saveInProgress;
}

this.restoreInProgress = function() {
    return _restoreInProgress;
}

const WORKSPACE_PREFS_NAME = "workspace2";

// This is ignored for now atleast until this code is made the defalt workspace code
const MOZ_PERSIST_POSITION_DOES_NOT_WORK =  platform.indexOf('linux') === 0;

const WORKSPACE_RESTORE_ON_START = "restoreOnRestart";
const WORKSPACE_ACTIVE_SPACE = "activeWorkspace";
const WORKSPACE_WINDOW_SETS = "workspaceWindows";
                               
/**
 * Restore workspace by name or workspace set to be recovered
 * at startup.
 *
 * This is the entry point for workspace recover from komodo.js
 * It is also implicitly used to restore windows when mutliple windows exist
 * in a workspace.
 */
this.restore = function ko_workspace_restore()
{
    try {
        // If there is only one window open and Komodo didn't shutdown cleanly.
        // Otherwise this was already asked in a previous iteration of restoring
        // a multi-window workspace.
        if (ko.windowManager.lastWindow() && ! _wasNormalShutdown()){
            log.debug("Komodo did not shutdown normally.  Asking to restore" +
                      "workspace.");
            if (! _askToRestore()) { // if no restore
                // Always restore the generic window state (separate from workspace prefs).
                // Must be called after the Mozilla persist state (onload) is done.
                setTimeout(ko.uilayout.restoreWindowState, 1);
                
                // Only restore the last main windows layout, eg id 1
                let prefs = ko.prefs;
                let path = []
                if (prefs.hasPref(WORKSPACE_PREFS_NAME)) {
                    let allWorkspacePrefs = prefs.getPref(WORKSPACE_PREFS_NAME);
                    if (allWorkspacePrefs.hasPref("default")) {
                        let workspacePrefs = allWorkspacePrefs.getPref("default");
                        if (workspacePrefs.hasPref("1")) {
                            prefs = workspacePrefs.getPref("1");
                            path = [WORKSPACE_PREFS_NAME,
                                    "default",
                                    "workspaceWindows",
                                    "1"];
                            ko.widgets.restoreLayout(prefs, path);
                            this.initializeEssentials(window);
                            log.debug("Not restoring workspace.");
                            return;
                        }
                    }
                } else {  // if workspace prefs just restore from globals
                    // Restore the default layout (i.e. last closed window)
                    ko.widgets.restoreLayout(ko.prefs, []);
                    return;
                }
            }
        }
        
        // This means we're window > 1 into a workspace restore.
        var workspacename = this.getNameFromWindow();
        this.restoreWorkspace(workspacename);
    } catch(ex) {
        log.exception(ex)
    } finally {
        this.initializeEssentials(window);
    }
}

/**
 * Check the window arguments for a workspace name and return it
 * 
 * @returns {string} name of workspace found 
 */
this.getNameFromWindow = function ko_workspace_getNameFromWindow() {
    if ('arguments' in window && window.arguments && window.arguments[0]) {
        var arg = window.arguments[0];
        if ('workspaceName' in arg) {
            return arg.workspaceName; // returns undefined if doesn't exist
        }
    }
}
 
/**
 * Restore a workspace.  Called by restore and open.
 * @param {String} A workspace name to recover.  Default is default ;)
 */
this.restoreWorkspace = function ko_workspace_restoreWorkspace(workspacename) {
    
    var allWorkspaces = null;
    //workspace to be restored
    var workspace = null;
    // always restore the default workspace for now unless specified by user
    // passing a name to restore or getting it from the window
    if(typeof(workspacename) == "undefined") {
        workspacename = "default";
    }
    log.debug("Workspace name: '" + workspacename + "'.");
    try { 
        _restoreInProgress = true;

        // Retrieve all workspaces
        if (ko.prefs.hasPref(WORKSPACE_PREFS_NAME)) {
            allWorkspaces = ko.prefs.getPref(WORKSPACE_PREFS_NAME);
        } else {
            log.warn("There are no workspace prefs to restore from.");
            setTimeout(ko.uilayout.restoreWindowState, 1);
            return;
        }
                
        if (allWorkspaces.hasPref(workspacename)) {
            log.debug("Found workspace " + workspacename + ".");
            workspace = allWorkspaces.getPref(workspacename);
        }
       
        if (workspace == null) {
            log.info("Workspace was null.  Nothing to recover.  Exiting");
            setTimeout(ko.uilayout.restoreWindowState, 1);
            return;
        }
        
        // Retrieve each window pref-set from the workspace and create a window
        // from that set
        var allWindowPrefs = workspace.getPref(WORKSPACE_WINDOW_SETS);
        var nextWindowId = this._getNextWindowIdToRestore(allWindowPrefs);
        var windowPrefSet = false;
        if (allWindowPrefs.hasPref(nextWindowId)) {
            windowPrefSet = allWindowPrefs.getPref(nextWindowId);
        }
        // We've already got a window to restore.
        var thisWindow = window;
        if ((typeof nextWindowId) != "undefined") {
            this.restoreWindow(thisWindow, windowPrefSet, workspacename);
        }
        
        // check if there is another window to restore
        if (this._getNextWindowIdToRestore(allWindowPrefs)) {
           ko.launch.newWindowFromWorkspaceName(workspacename);
        } else {
            // if it is undefined, we break out of the window generating loop
            // and we are now done restoring windows
            // Reset the restore state or else no windows will restore next time you
            // open this workspace
            this._resetWindowsRestoredState(allWindowPrefs);
        }
        
        // handle window.arguments spec list
        // This stuff will get moved to recovering files function
        // It will pull from preferences, not window args which workspace
        // shouldn't need to deal with any way.
        if ('arguments' in window && window.arguments && window.arguments[0]) {
            var arg = window.arguments[0];
            // There is no workspace to restore, but init window essentials
            this.initializeEssentials(window);
            var urllist;
            if ('uris' in arg) {
                urllist = arg.uris; // Called from ko.launch.newWindow(uri)
            } else if (arg instanceof Components.interfaces.nsIDialogParamBlock) {
                var paramBlock = arg.QueryInterface(Components.interfaces.nsIDialogParamBlock);
                urllist = paramBlock ? paramBlock.GetString(0).split('|') : [];
            } else if (typeof(arg) == 'string') {
                urllist = arg.split('|'); //see asCommandLineHandler.js
            } else {
                // arg is most likely an empty object
                urllist = [];
            }
            for (var i in urllist) {
                ko.open.URI(urllist[i]);
            }
        }
        // Probably pull this function.  Should only start Komodo from Default
        // Not using this function until Komodo uses workspace 2 on startup
        //this.setAsOnRestart(allWorkspaces, workspacename);
        // Some paths through the above block might not have called this,
        // so call it now to be sure.  See bug 87856 ericp.
        //  XXX careyh: Why call it at all on other code paths?  Just call it
        // when this code is done.
        this.initializeEssentials(thisWindow)
        _restoreFocusToMainWindow();
    } catch(ex) {
        log.exception(ex);
    } finally {
        _restoreInProgress = false;
    }
}

/**
 * Resets all windows in the workspace to a restore: false state when restore is
 * done
 */
this._resetWindowsRestoredState = function ko_workspace_resetWindowRestoredState(allWindowPrefs) {
    var prefIds = allWindowPrefs.getPrefIds();
    log.debug("Reseting all window prefs restored state to false.");
    var lim = prefIds.length;
    for (var i = 0; i < lim; i++) {
        var id = prefIds[i];
        var window = allWindowPrefs.getPref(id);
        if (window.hasBooleanPref("restored")) {
            window.setBooleanPref("restored", false);
        }
    }
};

this._getNextWindowIdToRestore = function ko_workspace_getNextWindowIdToRestore(allWindowPrefs) {
    
    // Fix up the stored window numbers
    let prefIds = allWindowPrefs.getPrefIds();
    prefIds = prefIds.map(function(n) parseInt(n, 10)).sort(function(a, b) a - b);
    if (prefIds[0] < 1) {
        // Invalid ids; shift everything over :|
        let prefs = prefIds.map(function(n) allWindowPrefs.getPref(n));
        //XXX this should be using Array.forEach.  Don't need a new array
        prefIds.map(function(n) allWindowPrefs.deletePref(n));
        for (let i = 1; prefs.length; ++i) {
            allWindowPrefs.setPref(i, prefs.shift());
        }
        prefIds = allWindowPrefs.getPrefIds();
    }
    //dump("_getNextWorkspaceIndexToRestore(" + currIdx +"): prefIds:" + prefIds + "\n");
    for (var i = 0; i < prefIds.length; i++) {
        var id = prefIds[i];
        var window = allWindowPrefs.getPref(id);
        if (window.hasBooleanPref("restored") && window.getBooleanPref("restored") == false) {
                return id;
            }
    }
    return undefined;
};
 
/**
 * restore all workspace preferences and state, open files and projects
 * @param {window} currentWindow, the window to perform restoration on
 * @param {preference set} the preferences to apply to the window
 * @param {string} workspacename, the workspace being restored.
 */
this.restoreWindow = function ko_workspace_restoreWindow(currentWindow,
                                                         windowPrefs,
                                                         workspacename)
{
    log.debug("Using window number " + currentWindow._koNum + " to restore.");
    // Always restore the generic window state (separate from workspace prefs).
    // Must be called after the Mozilla persist state (onload) is done.
    var wko = currentWindow.ko;
    // Restore the first workspace directly, and restore other
    // workspaces indirectly each new window's init routine in ko.main
    try {
        setTimeout(wko.uilayout.restoreWindowState, 1);
        if (windowPrefs.hasPref('coordinates')) {
            var coordinates = windowPrefs.getPref('coordinates');
            // Must be in a setTimeout, after the window has been loaded,
            // otherwise the window manager may resize or reposition it.
            setTimeout(_restoreWindowPosition, 1, currentWindow, coordinates);
        }
        
        // restore the views
        var id = "topview";
        var elt = currentWindow.document.getElementById(id);
        if (elt) {
            if (windowPrefs.hasPref(id)) {
                log.debug("Restoring views (files and their position/layout)");
                var pref = windowPrefs.getPref(id);
                log.debug("topview prefIds: " + pref.getPrefIds());
                elt.setState(pref);
            } else {
                log.debug("No previous view state saved.  Skipping.");
            }
        }
        
        try {
            log.debug("Restoring window layout");
            ko.widgets.restoreLayout(windowPrefs,
                                     [WORKSPACE_PREFS_NAME,
            /*eg. workspace name*/    windowPrefs.parent.id, 
                                      windowPrefs.id],
                                     true);
        } catch(e) {
            log.exception("Failed to restore window layout.  \
                          Restoring default.");
            ko.widgets.restoreLayout(ko.prefs, []);
        }
        
        
        if (wko.history) {
            log.debug("Restoring window history");
            wko.history.restore_prefs(windowPrefs);
        }
        // Projects depends on places, so open it after Places is initialized.
        if (windowPrefs.hasPref('opened_projects_v7')) {
            log.debug("Restoring Projects");
            let pref = windowPrefs.getPref('opened_projects_v7');
            var currentProjectURI;
            if (windowPrefs.hasPref('current_project')) {
                currentProjectURI = windowPrefs.getStringPref('current_project');
            } else {
                currentProjectURI = null;
            }
            // Don't load projects until places has initialized the projects view
            this.waitForProjectManager(function() {
                log.debug("Restoring project: " + currentProjectURI);
                wko.projects.manager.setState(pref);
                if (currentProjectURI) {
                    // If a project with that url is loaded, make it current
                    var proj = wko.projects.manager.getProjectByURL(currentProjectURI);
                    if (proj) {
                        wko.projects.manager.currentProject = proj;
                    }
                }
            });
        }
        wko._hasFocus = windowPrefs.getBoolean('hasFocus', false);
    } catch(ex) {
        log.exception(ex, "Error restoring workspace:");
    } finally {
        windowPrefs.setBooleanPref("restored", true);
    }
};

/**
 * Is Komodo in interactive mode?
 */
function _isNonInteractiveMode() {
    var infoSvc = Cc["@activestate.com/koInfoService;1"].getService();
    if (infoSvc.nonInteractiveMode) {
        return true;
    }
    return false;
}

/**
 * Ask user if they want to try restoring their workspace that has crashed
 */
function _askToRestore() {
    // If there is a workspace to restore - prompt the user to see if they wish
    // to restore it.
    if (ko.prefs.getBooleanPref("donotask_restore_workspace") &&
        ko.prefs.getStringPref("donotask_action_restore_workspace") == "No") {
        // The user has explictly asked never to restore the workspace.
        return false;
    }
    var prompt = _bundle.GetStringFromName("restoreWorkspaceAfterCrash.prompt");
    var title = _bundle.GetStringFromName("restoreWorkspaceAfterCrash.title");
    if (ko.dialogs.yesNo(prompt, null, null, title) == "No") {
        return false;
    } else if (ko.dialogs.yesNo(_bundle.GetStringFromName("doYouWantToOpenRecentFilesAndProjects.prompt"),
                                null, null, null, "restore_workspace") == "No") {
        return false;
    }
    return true;
}

/**
 * Check to see if Komodo shutdown normally
 * 
 * @returns {bool} true is Komodo shutdown normally
 */
function _wasNormalShutdown() {
    var was_normal_shutdown = ko.prefs.getBooleanPref('komodo_normal_shutdown');
    if (was_normal_shutdown) {
        ko.prefs.setBooleanPref('komodo_normal_shutdown', false);
        // Force flushing of prefs to file so pref isn't lost.
        var prefSvc = Cc["@activestate.com/koPrefService;1"].getService(Ci.koIPrefService);
        prefSvc.saveState();
    }
    return was_normal_shutdown;                                                 
}

/**
 * Take the set of all workspace prefs and returns the set with restoreOnRestart
 * set to true
 * @param {koPrefService} All workspace prefs
 */
this.getWorkspaceRestoreOnStart = function ko_workspace_getWorkspaceRestoreOnStart(allWorkspaces){
    var workspace = null;
    try {
        if (typeof allWorkspaces != "undefined") {
            var allWorkIDs = allWorkspaces.getPrefIds();
            for (let name of allWorkIDs) {
                workspace = allWorkspaces.getPref(name);
                if (workspace.hasPref(WORKSPACE_RESTORE_ON_START)) {
                    if (workspace.getPref(WORKSPACE_RESTORE_ON_START) == true) {
                        return workspace;
                    }
                } else {
                    log.info("No " + WORKSPACE_RESTORE_ON_START + "pref set on\
                             workspace " + name + ".");
                }
            }
        }
    } catch(e) {
        log.exception("Error while trying to load restoreOnStart workspace: " + e);
    } finally {
        return workspace;
    }
};


function _restoreFocusToMainWindow() {
    var windows = ko.windowManager.getWindows();
    for (var i = 0; i < windows.length; i++) {
        var w = windows[i];
        if (w.ko._hasFocus) {
            w.focus();
        }
        delete w.ko._hasFocus;
    }
}

// Bug 80604 -- screenX and screenY values like -32000 can occur.
// Generalize it: if we fall behind or in front of some threshold,
// return the acceptable min/max value.
function _checkWindowCoordinateBounds(candidateValue,
                            minAcceptableThreshold, minAcceptable,
                            maxAcceptableThreshold, maxAcceptable) {
    if (candidateValue < minAcceptableThreshold) {
        return Math.round(minAcceptable);
    }
    if (candidateValue > maxAcceptableThreshold) {
        return Math.round(maxAcceptable);
    }
    return candidateValue;
}

function _restoreWindowPosition(currentWindow, coordinates) {
    log.debug("Restoring window position.");
    const _nsIDOMChromeWindow = Components.interfaces.nsIDOMChromeWindow;
    var windowState = (coordinates.hasPrefHere('windowState')
                       ? coordinates.getLongPref('windowState')
                       : _nsIDOMChromeWindow.STATE_NORMAL);
    // If it's minimized or maximized we still need to set the
    // window's coords for when it's restored.
    var screenHeight = window.screen.availHeight;
    var screenWidth = window.screen.availWidth;
    var screenX = coordinates.getLongPref('screenX');
    var screenY = coordinates.getLongPref('screenY');
    var outerHeight = coordinates.getLongPref('outerHeight');
    var outerWidth = coordinates.getLongPref('outerWidth');
    if (Math.abs(screenX) > 3 * screenWidth || Math.abs(screenY) > 3 * screenHeight) {
        screenX = screenY = 0;
    }
    if (currentWindow.screenX != screenX || currentWindow.screenY != screenY) {
        currentWindow.moveTo(screenX, screenY);
    }
    if (currentWindow.outerHeight != outerHeight || currentWindow.outerWidth != outerWidth) {
        var newHeight = _checkWindowCoordinateBounds(outerHeight, 0,
                                                     .2 * screenHeight,
                                                     screenHeight,
                                                     .9 * screenHeight);
        var newWidth = _checkWindowCoordinateBounds(outerWidth, 0,
                                                    .2 * screenWidth,
                                                    screenWidth,
                                                    .9 * screenWidth);
        currentWindow.resizeTo(newWidth, newHeight);
    }
    if (windowState == _nsIDOMChromeWindow.STATE_MINIMIZED) {
        currentWindow.minimize();
    } else if (windowState == _nsIDOMChromeWindow.STATE_MAXIMIZED) {
        currentWindow.maximize();
    }
}

this.waitForProjectManager = function(callback) {
    // First make sure the places widget exists ,and then verify
    // the project manager has been hooked up, so the tree is loaded.
    ko.widgets.getWidgetAsync('placesViewbox', function() {
        var delayFunc;
        var limit = 100; // iterations
        var delay = 100;  // time in msec
        delayFunc = function(tryNum) {
            try {
                if (ko.toolbox2 && ko.toolbox2.manager &&
                    ko.projects.manager.viewMgr.owner.projectsTreeView) {
                    callback();
                    return;
                }
            } catch(ex) {
                log.info("waitForProjectManager: Failure: " + tryNum + ": "  + ex);
            };
            if (tryNum < limit) {
                setTimeout(delayFunc, delay, tryNum + 1);
            } else {
                log.error("waitForProjectManager: Gave up trying to restore the projects workspace");
            }
        }
        setTimeout(delayFunc, delay, 0);
    });
};

this._calledInitializeEssentials = false;

this.initializeEssentials = function(currentWindow) {
    if (this._calledInitializeEssentials) {
        return;
    }
    var infoService = Components.classes["@activestate.com/koInfoService;1"].
                                 getService(Components.interfaces.koIInfoService);
    if (!("__koNum" in currentWindow.ko.main)) {
        currentWindow._koNum = infoService.nextWindowNum();
    }
    xtk.domutils.fireEvent(window, 'workspace_restored');
    this._calledInitializeEssentials = true;
}


// write the current state of preferences to disk
this._flushPreferences = function ko_workspace__dumpPreferences(saveNow){
    var prefSvc = Cc["@activestate.com/koPrefService;1"].getService(Ci.koIPrefService);
    try{
        if (saveNow) {
            prefSvc.saveState();
        } else {
            prefSvc.saveWhenIdle();
        }
    } catch(e) {
        log.exception(e,"Error writing preferences to disk: ");
    }
}

/**
 * Save in an MRU for user restoring operation.
 */
this.addToWorkspaceMRU = function ko_workspace_addToWorkspaceMRU(workspaceName) {
    
    if (!ko.prefs.hasPref("mruWorkspaceSize")) {
        ko.prefs.setLong("mruWorkspaceSize", 10);
    }
    ko.mru.add("mruWorkspaceList", workspaceName);
}

/**
 * Get prefs set if it exists.  Returns set if it exists. Otherwise returns null
 * @param {String} name of set
 * @param {prefs object} all workspace preferences
 */
this.getExistingWorkspaceSet = function ko_workspace_getExistingWorkspaceSet(workspacePrefs, name) {
    if (workspacePrefs.hasPref(name)) {
        log.debug('Found existing workspace prefs for ' + name);
        return workspacePrefs.getPref(name);
    }
    log.debug('No existing workspace prefs for ' + name);
    return null;
}

/**
 * Delete the entire Workspace pref namespace and anything in it
 * NOTE currently only for testing.  This function shouldn't be necessary in
 * the wild.
 */
this.deleteAllWorkspacePrefs = function ko_workspace_deleteAllPrefs() {
    ko.pref.deletePref(WORKSPACE_PREFS_NAME);
}

/**
 * Save the current workspace (All open windows, each windows state, all open 
 * files in each window).  If a pref set exists with the same name, it is over
 * written with the new set. eg. 'default' is over written everytime this is run
 * with no argument.
 *
 * This function also automatically populates an MRU for workspaces
 * 
 * @public
 * @param {String} name of the space to save
 */
this.saveWorkspace = function ko_workspace_saveWorkspace(name /*=default*/) {
    _saveInProgress = true;
    name = typeof name !== 'undefined' ? name : 'default';
    log.debug('Start workspace save for workspace "' + name + '".')
    // Global workspace prefs, eg.  all of them
    var allWorkspaces = this._getBaseWorkspacePrefSet();
    
    // Check if there is an existing set of prefs with this name and ask user
    // if they want to over-write them.
    var workspacePrefs = this.getExistingWorkspaceSet(allWorkspaces, name);
    // If no exiting workspace is found then make a new entry
    if (workspacePrefs == null) {
        workspacePrefs = Cc['@activestate.com/koPreferenceSet;1'].createInstance();
    }
    allWorkspaces.setPref(name, workspacePrefs);
    // Grab all the windows to loop through
    var openWindows = ko.windowManager.getWindows();
    // Pref set to save windows under
    var windows = Cc['@activestate.com/koPreferenceSet;1'].createInstance();
    workspacePrefs.setPref(WORKSPACE_WINDOW_SETS, windows);
    try {
        for(let i in openWindows){
            let thisWindow = openWindows[i];
            this.saveWindowsState(thisWindow, windows, name);
        }  
        // Divide the # of millisec by 1000, or we'll overflow on the setLongPref
        // conversion to an int.  Legacy
        workspacePrefs.setLongPref('timestamp', (new Date()).valueOf() / 1000);
        
        // Add the new workspace pref set to the global set
        this.addToWorkspaceMRU(name);
        // Because if it's being saved that means it's active, even at shutdown
        this.setAsOnRestart(allWorkspaces, name);
        //cleanse workspace prefs and MRU
        //this.cleanseWorkspacePrefsMRU();
        // Save prefs state to disk
        this._flushPreferences();
        log.debug("Workspace '" + name + "' has been saved.  Enjoy!");
    } catch(e) {
        log.exception(e,"Error saving workspace: ");
    } finally {
        _saveInProgress = false;
    }
}

/**
 * Set the WORKSPACE_RESTORE_ON_START pref on the currently active Prefset
 * Make sure it's disabled in all other workspaces
 *
 * @param {koPrefService}  All workspace pref sets
 * @param {String} Workspace to be set as WORKSPACE_RESTORE_ON_START
 */
this.setAsOnRestart = function ko_workspace_setOnRestart(allWorkspaces, name) {
    // Set named set as restore set
    if (allWorkspaces.hasPref(name)) {
        let workspacePrefs = allWorkspaces.getPref(name);
        workspacePrefs.setBooleanPref(WORKSPACE_RESTORE_ON_START, true);
    } else {
        log.warn("Could not set Workspace Restore pref on " + name +".");
    }
    // set all other workspaces to false
    var allWorkIDs = allWorkspaces.getPrefIds();
    for (let name of allWorkIDs) {
        if (allWorkspaces.hasPref(name)) {
            let workspace = allWorkspaces.getPref(name);
            workspace.setBooleanPref(WORKSPACE_RESTORE_ON_START, false);
        }
    }
}

/**
 * Returns a preference object with the windows state.
 * Saves window location, size, current project, open files, etc.
 *
 * @param {Window} Takes a Komodo window object
 */
this.saveWindowsState = function ko_workspace_saveWindowsState(thisWindow,
                                                               windowPrefs,
                                                               workspacename) {
    log.debug('Saving Window' + thisWindow._koNum + ' State.');
   //Create a pref object to save the window state in
    var windowWorkspace =
        Components.classes['@activestate.com/koPreferenceSet;1'].createInstance();
    windowPrefs.setPref(thisWindow._koNum, windowWorkspace);

    // State collecting and appending state information to the window state pref
    log.debug('\tSaving Window coordinates.');
    var coordinates = this.getWindowCoordinateState(thisWindow);
    windowWorkspace.setPref('coordinates', coordinates);
    
    log.debug('\tSaving project state.');
    this.getWindowProjectState(thisWindow, windowWorkspace);
    
    log.debug('\tSaving view state.');
    var viewPrefs = this.getWindowViewState(thisWindow);
    if (viewPrefs) {
        windowWorkspace.setPref(viewPrefs.id, viewPrefs);
    }
    
    log.debug('\tSaving open files.');
    // must call ko from the window being saved as 
    var openfilesPrefs = this.collectOpenViewPaths(thisWindow);
    if (openfilesPrefs.length > 0) {    
        windowWorkspace.setPref("openfiles", openfilesPrefs);
    } else {
        //XXX restore needs to accommodate this so it doesn't fail when it
        // can't find any files to restore
        log.debug('\tNo view open.  No view state saved on window ' +
                  thisWindow._koNum + '.');
    }
    
    // Save the window widget layout state
    thisWindow.ko.widgets.unload([WORKSPACE_PREFS_NAME,
                                  workspacename,
                                  WORKSPACE_WINDOW_SETS,
                                  thisWindow._koNum]);
    
    windowWorkspace.setLongPref('windowNum', thisWindow._koNum);
    // Save the current windows ko.history to the window workspace
    if (thisWindow.history) {
        log.debug('\tSaving window history.');
        thisWindow.ko.history.save_prefs(windowWorkspace);
    }
    
    // Set restored state for restoration process
    windowWorkspace.setBooleanPref("restored", false);
    log.debug('Window ' + thisWindow._koNum + '  state save:  COMPLETE');
};

/**
 * Return a preference object with the views state in it
 * @param {Window} A Komodo Window
 */
this.getWindowViewState = function ko_workspace_getWindowViewState(thisWindow) {
    var pref;
    var id = 'topview';
    var elt;
    elt = thisWindow.document.getElementById(id);
    if (!elt) {
        log.info(_bundle.formatStringFromName("couldNotFind.alert", [id], 1));
        return;
    }
    var pref = elt.getState();
    if (pref) {
        pref.id = id;
    }
    return pref;
}

/**
 * Return a preference object with the Projects state in it
 * @param {Window} A Komodo Window
 * @param {prefset} the window prefs set being saved
 */
this.getWindowProjectState = function ko_workspace_getWindowOpenProject(thisWindow, prefs) {
    var wko = thisWindow.ko;
    var pref = wko.projects.manager.getState();
    if (pref) {
        prefs.setPref(pref.id, pref);
        var currentProject = wko.projects.manager.currentProject;
        if (currentProject) {
            prefs.setStringPref('current_project', currentProject.url);
        }
    }
}

/**
 * Returns a preferences object with the windows coordinates 
 * @param {Window} A Komodo Window
 */
this.getWindowCoordinateState = function ko_workspace_getWindowCoordinateState(thisWindow) {
    var coordinates =
        Components.classes['@activestate.com/koPreferenceSet;1'].createInstance();
    
    coordinates.setLongPref('windowState', thisWindow.windowState);
    var docElement = thisWindow.document.documentElement;
    coordinates.setLongPref('screenX', docElement.getAttribute('screenX'));
    coordinates.setLongPref('screenY', docElement.getAttribute('screenY'));
    coordinates.setLongPref('outerHeight', docElement.height);
    coordinates.setLongPref('outerWidth', docElement.width);
    
    return coordinates;
}

/**
 * Creates/returns the base workspace pref object that will contain all
 * saved workspaces
 */
this._getBaseWorkspacePrefSet = function() {
    if (ko.prefs.hasPref(WORKSPACE_PREFS_NAME)) {
        return ko.prefs.getPref(WORKSPACE_PREFS_NAME);
    }
    var workspacePrefs = Components.classes['@activestate.com/koPreferenceSet;1'].createInstance();
    ko.prefs.setPref(WORKSPACE_PREFS_NAME, workspacePrefs);
    return workspacePrefs;
}

/**
* Collect path to all open files
* @returns (Array) all open file paths
*/
this.collectOpenViewPaths = function(){
    var openFilePrefs = Components.classes["@activestate.com/koOrderedPreference;1"].
            createInstance(Components.interfaces.koIOrderedPreference);
    var curViews = ko.views.manager.getAllViews();
    for (var i = 0; i < curViews.length; i++){
        var currentView = curViews[i];
        // don't want browser views
        if (currentView.getAttribute("type") != "editor") {
            continue;
        }
        // Don't want macros or snippets user might be editing
        // XXX Could provide an option in the workspace prfs later to include
        if (!currentView.getURI().startsWith("file")) {
            continue;
        }
        var uri = false;
        try {
            uri = currentView.getURI()
        } catch (ex) {
            log.warn("Could not retrieve URI from view: " + ex );
            try {
                uri = ko.uriparse.pathToURI(currentView.koDoc.file.path);
                // Might as well skip this loop with no file path
                continue;
            } catch (ex) {
                log.warn("Can't retrieve path from view: " + ex);
            }
        }
        if (uri) {
            openFilePrefs.appendStringPref(uri);
        }
    }
    return openFilePrefs;
}


/**
* UI API entry to save workspace
*/
this.save = function ko_workspace_save(workspacename){
    if (typeof(workspacename) ==  "undefined") {
        let yesNoMsg = _bundle.GetStringFromName("saveWorkspaceSaveNew.YesNoCancel");
        var saveNewWorkspace = ko.dialogs.yesNoCancel(yesNoMsg);
        var workSpaceMRUList = ko.mru.getAll("mruWorkspaceList");

        if (saveNewWorkspace == "Yes") {
            var message = _bundle.GetStringFromName("saveWorkspaceSaveNew.message");  
            var opts = {
                label: _bundle.GetStringFromName("saveWorkspaceSaveNew.label"),
                value: _bundle.GetStringFromName("saveWorkspaceSaveNew.value"),
                title: _bundle.GetStringFromName("saveWorkspaceSaveNew.title"),
                mruName: _bundle.GetStringFromName("saveWorkspaceSaveNew.mruname")
            }
            workspacename = require("ko/dialogs").prompt(message,opts);
            // Check if the name exists and prompt user to re-enter name
            // Keep promping if they put in another existing name
            var sameName = true;
            while (sameName) {
                if (workSpaceMRUList.indexOf(workspacename) != -1) {
                    let nameExistsMsg = _bundle.GetStringFromName("saveWorkspaceSaveNewExists.YesNoCancel");
                    var nameExistsCont = ko.dialogs.yesNoCancel(nameExistsMsg);
                    if (nameExistsCont == "Yes") {
                        sameName = false;
                    } else if (nameExistsCont == "No") {
                        workspacename = require("ko/dialogs").prompt(message,opts);
                    } else {
                        // this means they cancelled
                        workspacename = null;
                        sameName = false;
                    }
                } else {
                    sameName = false;
                }
            }
            
        } else {
            let title = _bundle.GetStringFromName("saveWorkspaceSaveExisting.title");
            let prompt = _bundle.GetStringFromName("saveWorkspaceSaveExisting.prompt");
            let save = _bundle.GetStringFromName("saveWorkspaceSaveExisting.save");
            let cancel = _bundle.GetStringFromName("saveWorkspaceSaveExisting.cancel");

            workspacename = ko.dialogs.selectFromList(title,
                                             prompt,
                                             workSpaceMRUList,
                                             "one", /*selectionCondition*/
                                             null, /*stringifier*/
                                             null, /*doNotAskPref*/
                                             false, /*yesNoCancel*/
                                             [save,cancel], /*buttons*/
                                             0);   /*selectedIndex*/
        }
    }
    // null is returned from both dialogs if cancel is clicked.  Better not save
    if (!workspacename) {
        return;
    }
    log.debug("saveing worksapce with name: " + workspacename  );
    try {
        this.saveWorkspace(workspacename);
    } catch(e) {
        log.exception("An error occurred while saving workspace. ERROR: " + e);
    }
}

/**
* Take a list of URIs and load the files into Komodo based on those paths.
* @params {String} name of workspace to restore
*/
this.open = function(workspacename) {
    if (typeof(workspacename) ==  "undefined") {
        var workSpaceMRUList = ko.mru.getAll("mruWorkspaceList");
        let title = _bundle.GetStringFromName("restorWorkspaceOpen.title");
        let prompt = _bundle.GetStringFromName("restorWorkspaceOpen.prompt");
        let open = _bundle.GetStringFromName("restorWorkspaceOpen.open");
        let cancel = _bundle.GetStringFromName("restorWorkspaceOpen.cancel");

        workspacename = ko.dialogs.selectFromList(title,
                                         prompt,
                                         workSpaceMRUList,
                                         "one", /*selectionCondition*/
                                         null, /*stringifier*/
                                         null, /*doNotAskPref*/
                                         false, /*yesNoCancel*/
                                         [open,cancel], /*buttons*/
                                         0);   /*selectedIndex*/
    }
    
    if (!workspacename) {
        // Must have cancelled.  Fine...i'll just do nothing then!
        return;
    }    
    try {
        // If there are other Windows open, tell user to close all but one and try
        // again.
        if (ko.windowManager.getWindows().length > 1) {
            ko.dialogs.alert(_bundle.GetStringFromName("restoreWorkspaceCloseWindows.text"));
            return;
        }
        this.restoreWorkspace(workspacename);
    } catch(e) {
        log.exception("Failed to load workspace:  ERROR: " + e)
    }
}

}).apply(ko.workspace2);
